import importlib
import sys
from uvicore.typing import Any, Dict, List, Union, Dict

import uvicore
#import typer_async as typer
from uvicore.contracts import Application, Package
from uvicore.contracts import Provider as ProviderInterface
from uvicore.contracts import Dispatcher
from uvicore.console import group as click_group
from uvicore.support.dumper import dd, dump
from uvicore.support.module import load, location


@uvicore.service(aliases=['PackageProvider', 'Provider'])
class Provider(ProviderInterface):

    # Class variable of all registered click groups from app packages
    # Used to extend an existing group from other packages
    __click_groups__ = {}

    @property
    def app(self) -> Application:
        # """Uvicore application instance"""
        return self._app

    @property
    def events(self) -> Dispatcher:
        # """Event instance"""
        return uvicore.events

    @property
    def package(self) -> Package:
        # """The current package class.  Not available in boot()"""
        return self._package

    @property
    def app_config(self) -> Dict:
        return self._app_config

    @property
    def package_config(self) -> Dict:
        return self._package_config

    @property
    def name(self) -> str:
        return self._name

    def __init__(self, app: Application, name: str, package: Package, app_config: Dict, package_config: Dict) -> None:
        self._app = app
        self._name = name
        self._package = package
        self._app_config = app_config
        self._package_config = package_config

    def bind(self,
        name: str,
        object: Any,
        *,
        object_type: str = 'service',
        override: bool = True,
        factory: Any = None,
        kwargs: Dict = None,
        singleton: bool = False,
        aliases: List = []
    ) -> None:
        """Bind objects to the Ioc"""
        # Get override object from config if exists
        override = self.binding(name)
        object = override or object

        # Bind object to IoC
        uvicore.ioc.bind(name, object, factory=factory, kwargs=kwargs, singleton=singleton, aliases=aliases)

    def bind_override(self, name: str, object: str):
        uvicore.ioc.bind_override(name, object)

    def binding(self, name: str) -> str:
        if self.app_config.get('bindings'):
            return self.app_config.get('bindings').get(name)

    def configs(self, options: List[Dict]) -> None:
        for config in options:
            value = None;
            if 'module' in config:
                # Load module to get actual config value
                value = load(config['module']).object
            elif 'value' in config:
                # Actual config already passed
                value = config['value']

            # Don't use is None, because empty Dict({}) won't be caught, must use if not value
            if not value: continue;

            # If uvicore.config is None (brand new) set to empty Dict() before .merge()
            if uvicore.config is None: uvicore.config = Dict()

            # Merge config value with complete config
            new = Dict()
            new.dotset(config['key'], value)
            uvicore.config.merge(new)

    def registers(self, options: Dict) -> None:
        if options is not None:
            self.package.registers = Dict(options)


# IoC Class Instance
#_ServiceProviderIoc: _ServiceProvider = uvicore.ioc.make('ServiceProvider', _ServiceProvider, aliases=['service', 'provider'])

#class ServiceProvider(_ServiceProviderIoc, ProviderInterface):
#    pass

# Public API for import * and doc gens
#__all__ = ['ServiceProvider', '_ServiceProvider']



from __future__ import annotations

import logging
import logging.config
import re
import sys
from logging import Formatter
from logging import Logger as PythonLogger

from colored import attr, bg, fg

import uvicore
from uvicore.contracts import Logger as LoggerInterface
from uvicore.support.dumper import log_dump

# # Sunfinity standardized log configuration
# config = {
#     'version': 1,
#     'formatters': {
#         'console': {
#             'format': '%(message)s',
#             'datefmt': '%Y-%m-%d %H:%M:%S',
#         },
#         'file': {
#             #format': ''%(asctime)s.%(msecs)03d | %(process)-5d | %(levelname)-7s | %(module)s:%(lineno)-10d | %(message)s'',
#             #format': ''%(asctime)s.%(msecs)03d | %(levelname)-7s | %(message)s'',
#             'format': '%(asctime)s.%(msecs)03d | %(levelname)-7s | %(message)s',
#             'datefmt': '%Y-%m-%d %H:%M:%S',
#         },
#     },
#     'handlers': {
#         'console': {
#             'class': 'logging.StreamHandler',
#             'level': 'INFO',
#             'formatter': 'console',
#             'stream': 'ext://sys.stdout',
#         },
#         'file': {
#             'class': 'logging.handlers.RotatingFileHandler',
#             'level': 'DEBUG',
#             'formatter': 'file',
#             'filename': file,
#             'maxBytes': 1024,
#             'backupCount': 3,
#         },
#     },
#     'root': {
#         'level': 'DEBUG',
#         'handlers': ['console', 'file'],
#     }
# }


class OutputFilter(logging.Filter):
    """Python logging custom filter class"""

    def __init__(self, filters, excludes):
        self.filters = filters
        self.excludes = excludes
        super().__init__(name='')

    def filter(self, record):
        # Not an exact filter match but a contains match.  This matches how default python
        # logging filters are.  So you can filter on A.B and it will include
        # names of A.B.C and up.
        show = False
        if self.filters:
            for f in self.filters:
                if record.name[0:len(f)] == f:
                    show = True
                    break
        else:
            show = True

        if show and self.excludes:
            for exclude in self.excludes:
                if record.name[0:len(exclude)] == exclude:
                    show = False
                    break

        return show



class ExcludeFilter(logging.Filter):
    """Python logging custom exclude filter class"""

    def __init__(self, excludes):
        self.excludes = excludes
        super().__init__(name='exclude')

    def filter(self, record):
        # Not an exact filter match but a contains match.  This matches how default python
        # logging filters are.  So you can filter on A.B and it will include
        # names of A.B.C and up.
        for exclude in self.excludes:
            print(exclude)
            if record.name[0:len(exclude)] == exclude: return False
        return True


class ColoredFormatter(Formatter):

    def __init__(self, patern):
        Formatter.__init__(self, patern)

    def format(self, record):
        # Remember this is console output only, not file or other handlers
        # See color chart https://pypi.org/project/colored/
        level = record.levelname
        message = logging.Formatter.format(self, record)

        # Format header
        if (level == 'INFO' and re.match("^:: ", message)):
            message = re.sub("^:: ", "", message)
            message = re.sub(" ::$", "", message)
            message = ('{0}{1}{2}{3}').format(fg('dark_orange'), attr('bold'), ':: ', attr(0)) \
                + ('{0}{1}{2}{3}').format(fg('green'), attr('bold'), message, attr(0)) \
                + ('{0}{1}{2}{3}').format(fg('dark_orange'), attr('bold'), ' ::', attr(0))

        # Format header2
        if (level == 'INFO' and re.match("^## ", message)):
            message = re.sub("^## ", "", message)
            message = re.sub(" ##$", "", message)
            message = ('{0}{1}{2}{3}').format(fg('dark_orange'), attr('bold'), '## ', attr(0)) \
                + ('{0}{1}{2}{3}').format(fg('green'), attr('bold'), message, attr(0)) \
                + ('{0}{1}{2}{3}').format(fg('dark_orange'), attr('bold'), ' ##', attr(0))

        # Format header3
        if (level == 'INFO' and re.match("^=== ", message)):
            message = re.sub("^=== ", "", message)
            message = re.sub(" ===$", "", message)
            message = ('{0}{1}{2}{3}').format(fg('dark_orange'), attr('bold'), '=== ', attr(0)) \
                + ('{0}{1}{2}').format(fg('green'), message, attr(0)) \
                + ('{0}{1}{2}{3}').format(fg('dark_orange'), attr('bold'), ' ===', attr(0))

        # Format header4
        if (level == 'INFO' and re.match("^---- ", message)):
            message = re.sub("^---- ", "", message)
            message = re.sub(" ----$", "", message)
            message = ('{0}{1}{2}{3}').format(fg('dark_orange'), attr('bold'), '---- ', attr(0)) \
                + ('{0}{1}{2}').format(fg('dark_green'), message, attr(0)) \
                + ('{0}{1}{2}{3}').format(fg('dark_orange'), attr('bold'), ' ----', attr(0))

        # Format bullet * item
        elif (level == 'INFO' and re.match("^\* ", message)):
            message = re.sub("^\* ", "", message)
            message = ('{0}{1}{2}').format(fg('green'), '   * ', attr(0)) \
                + ('{0}{1}{2}{3}').format(fg('white'), attr('bold'), message, attr(0))

        # Format bullet2 - item
        elif (level == 'INFO' and re.match("^- ", message)):
            message = re.sub("^- ", "", message)
            message = ('{0}{1}{2}').format(fg('cyan'), '   - ', attr(0)) \
                + ('{0}{1}{2}{3}').format(fg('white'), attr('bold'), message, attr(0))

        # Format bullet3 + item
        elif (level == 'INFO' and re.match("^\+ ", message)):
            message = re.sub("^\+ ", "", message)
            message = ('{0}{1}{2}').format(fg('red'), '   + ', attr(0)) \
                + ('{0}{1}{2}{3}').format(fg('white'), attr('bold'), message, attr(0))

        # Format bullet4 > item
        elif (level == 'INFO' and re.match("^> ", message)):
            message = re.sub("^> ", "", message)
            message = ('{0}{1}{2}').format(fg('magenta'), '   > ', attr(0)) \
                + ('{0}{1}{2}{3}').format(fg('white'), attr('bold'), message, attr(0))

        # Format notice
        elif (level == 'INFO' and re.match("^NOTICE: ", message)):
            message = re.sub("^NOTICE: ", "", message)
            message = ('{0}{1}{2}{3}').format(fg('yellow'), attr('bold'), 'NOTICE: ', attr(0)) \
                + ('{0}{1}{2}{3}').format(fg('white'), attr('bold'), message, attr(0))

        # Format separator
        elif (level == 'INFO' and re.match("^====", message)):
            message = ('{0}{1}{2}{3}').format(fg('orange_4a'), attr('bold'), message, attr(0))

        # Format line
        elif (level == 'INFO' and re.match("^----", message)):
            message = ('{0}{1}{2}{3}').format(fg('orange_4a'), attr('bold'), message, attr(0))

        elif (level == 'DEBUG'):
            message = ('{0}{1}{2}').format(fg(241), message, attr(0))
        elif (level == 'INFO'):
            message = message
        elif (level == 'WARNING'):
            message = ('{0}{1}{2}').format(fg('orange_red_1'), message, attr(0))
        elif (level == 'ERROR'):
            message = ('{0}{1}{2}').format(fg('red'), message, attr(0))
        elif (level == 'CRITICAL'):
            message = ('{0}{1}{2}{3}').format(fg('black'), bg('red'), message, attr(0))

        return message


@uvicore.service('uvicore.logging.logger.Logger',
    aliases=['Logger', 'logger', 'Log', 'log'],
    singleton=True,
    kwargs={'config': uvicore.config('app.logger')},
)
class Logger(LoggerInterface):
    """Logger private class.

    Do not import from this location.
    Use the uvicore.log singleton global instead."""

    def __init__(self, config):
        # Default Config
        # Levels from logging._levelToName are
        # {50: 'CRITICAL', 40: 'ERROR', 30: 'WARNING', 20: 'INFO', 10: 'DEBUG', 0: 'NOTSET'}

        # Levels = DEBUG, INFO, WARNING, ERROR, CRITICAL
        default = {
            'console': {
                'enabled': True,
                'level': 'DEBUG',
                'colors': True,
                'format': '%(message)s',
                'filters': [],
                'exclude': [],
            },
            'file': {
                'enabled': False,
                'level': 'DEBUG',
                'file': '/tmp/example.log',
                'when': 'midnight',
                'interval': 1,
                'backup_count': 7,
                'format': '%(asctime)s.%(msecs)03d | %(levelname)-8s | %(name)-22s | %(message)s',
                'filters': [],
                'exclude': [],
            }
        }

        # Merge default and user defined config
        config = {**default, **config}
        if 'console' in config.keys(): config['console'] = {**default['console'], **config['console']}
        if 'file' in config.keys(): config['file'] = {**default['file'], **config['file']}

        # New Logger
        self._logger = logging.getLogger()
        self._logger.setLevel(logging.DEBUG)
        self._name = None

        # New Console Handler
        if config['console']['enabled']:
            handler = logging.StreamHandler(stream=sys.stdout)
            handler.setLevel(config['console']['level'])
            if config['console']['colors']:
                handler.setFormatter(ColoredFormatter(config['console']['format']))
            else:
                handler.setFormatter(logging.Formatter(
                    fmt=config['console']['format'],
                    datefmt='%Y-%m-%d %H:%M:%S'
                ))
            handler.addFilter(OutputFilter(config['console']['filters'], config['console']['exclude']))
            self._logger.addHandler(handler)

        # New File Handler
        if config['file']['enabled']:
            #class logging.handlers.TimedRotatingFileHandler(filename, when='h', interval=1, backupCount=0, encoding=None, delay=False, utc=False, atTime=None, errors=None)
            #handler = logging.FileHandler(filename=config['file']['file'], mode='a')
            handler = logging.handlers.TimedRotatingFileHandler(filename=config['file']['file'], when=config['file']['when'], interval=config['file']['interval'], backupCount=config['file']['backup_count'])
            handler.setLevel(config['file']['level'])
            handler.setFormatter(logging.Formatter(
                fmt=config['file']['format'],
                datefmt='%Y-%m-%d %H:%M:%S'
            ))
            #if config['file'].get('filter'): handler.addFilter(logging.Filter(name=config['file']['filter']))
            handler.addFilter(OutputFilter(config['file']['filters'], config['file']['exclude']))
            self._logger.addHandler(handler)

        self.config = config

    def __call__(self, message):
        self.info(message)

    @property
    def console_handler(self) -> PythonLogger:
        try:
            return self._logger.handlers[0]
        except IndexError:
            return None

    @property
    def file_handler(self) -> PythonLogger:
        try:
            return self._logger.handlers[1]
        except IndexError:
            return None

    @property
    def logger(self):
        if not self._name: return self._logger
        return logging.getLogger(self._name)

    def name(self, name: str):
        self._name = name
        return self

    def reset(self):
        self._name = None

    def dump(self, *args):
        self._dump_handler(*args, handler='file', filters=self.config['file']['filters'], excludes=self.config['file']['exclude'])
        self._dump_handler(*args, handler='console', filters=self.config['console']['filters'], excludes=self.config['console']['exclude'])
        self.reset()

    def _dump_handler(self, *args, handler: str, filters: List, excludes: List):
        if not self._name: self.name == 'root'
        show = False

        # Check filters
        if not filters: show = True
        if not show and self._name:
            for filter in filters:
                if self._name[0:len(filter)] == filter:
                    show = True
                    break

        # Check excludes
        if show and excludes and self._name is not None:
            for exclude in excludes:
                if self._name[0:len(exclude)] == exclude:
                    show = False
                    break

        if show:
            if handler == 'console':
                # Pretty Printer to Console
                log_dump(*args)
            else:
                # Print as text to file handler
                for arg in args:
                    self.logger.debug(arg)

    def info(self, message):
        self.logger.info(str(message))
        self.reset()

    def notice(self, message):
        self.logger.info("NOTICE: " + str(message))
        self.reset()

    def warning(self, message):
        self.logger.warning(str(message))
        self.reset()

    def debug(self, message):
        self.logger.debug(str(message))
        self.reset()

    def error(self, message):
        self.logger.error(str(message))
        self.reset()

    def critical(self, message):
        self.logger.critical(str(message))
        self.reset()

    def exception(self, message):
        self.logger.error(str(message))
        self.reset()

    def blank(self) -> LoggerInterface:
        self.logger.info('')
        self.reset()

    def nl(self) -> LoggerInterface:
        """nl() is a blank() that is chainable"""
        self.logger.info('')
        return self

    def separator(self):
        self.logger.info('=' * 80)
        self.reset()

    def line(self):
        self.logger.info('-' * 80)
        self.reset()

    def header(self, message):
        self.logger.info(":: " + str(message) + " ::")
        self.reset()

    def header2(self, message):
        self.logger.info("## " + str(message) + " ##")
        self.reset()

    def header3(self, message):
        self.logger.info("=== " + str(message) + " ===")
        self.reset()

    def header4(self, message):
        self.logger.info("---- " + str(message) + " ----")
        self.reset()

    def item(self, message):
        self.logger.info("* " + str(message))
        self.reset()

    def item2(self, message):
        self.logger.info("- " + str(message))
        self.reset()

    def item3(self, message):
        self.logger.info("+ " + str(message))
        self.reset()

    def item4(self, message):
        self.logger.info("> " + str(message))
        self.reset()



# IoC Class Instance
# Not to be imported by the public from here.
# Use the uvicore.log singleton global instead.

# Public API for import * and doc gens
#__all__ = ['_Logger', 'ColoredFormatter']



from __future__ import annotations
from abc import ABC, abstractmethod, abstractproperty
from typing import Dict, Any
from logging import Logger as PythonLogger

class Logger(ABC):

    @property
    @abstractmethod
    def console_handler(self) -> PythonLogger:
        """Get the console logger"""

    @abstractproperty
    @abstractmethod
    def file_handler(self) -> PythonLogger:
        """Get the file logger"""

    @abstractproperty
    @abstractmethod
    def logger(self):
        """Get the logger"""

    @abstractmethod
    def name(self, name: str):
        """Set name of logger"""

    @abstractmethod
    def reset(self):
        """Clear logger name"""

    @abstractmethod
    def dump(self, *args):
        """Dump message"""

    @abstractmethod
    def info(self, message) -> Logger:
        """Log an info message"""

    @abstractmethod
    def notice(self, message) -> Logger:
        """Log a notice message"""

    @abstractmethod
    def warning(self, message) -> Logger:
        """Log a warning message"""

    @abstractmethod
    def debug(self, message) -> Logger:
        """Log a debug message"""

    @abstractmethod
    def error(self, message) -> Logger:
        """Log an error message"""

    @abstractmethod
    def critical(self, message) -> Logger:
        """Log a critical message"""

    @abstractmethod
    def exception(self, message) -> Logger:
        """Log an exception message"""

    @abstractmethod
    def blank(self) -> Logger:
        """Log a blank line"""

    @abstractmethod
    def nl(self) -> Logger:
        """Log a blank line"""

    @abstractmethod
    def separator(self) -> Logger:
        """Log a = line separator"""

    @abstractmethod
    def line(self) -> Logger:
        """Log a - line separator"""

    @abstractmethod
    def header(self, message) -> Logger:
        """Header :: style"""

    @abstractmethod
    def header2(self, message) -> Logger:
        """Header ## style"""

    @abstractmethod
    def header3(self, message) -> Logger:
        """Header === style"""

    @abstractmethod
    def header4(self, message) -> Logger:
        """Header ---- style"""

    @abstractmethod
    def item(self, message) -> Logger:
        """Item * style"""

    @abstractmethod
    def item2(self, message) -> Logger:
        """Item - style"""

    @abstractmethod
    def item3(self, message) -> Logger:
        """Item + style"""

    @abstractmethod
    def item4(self, message) -> Logger:
        """Item > style"""



from __future__ import annotations
import uvicore
from uvicore.typing import Any, Callable, List, Dict, Optional, Decorator, get_type_hints, Union, Type
from uvicore.http.routing.router import Router
from uvicore.contracts import ApiRoute as RouteInterface
from prettyprinter import pretty_call, register_pretty
from uvicore.support.dumper import dump, dd
from merge_args import _merge as merge_args
from functools import partial
from uvicore.http.response import Response, JSON as JSONResponse
from fastapi.datastructures import Default, DefaultPlaceholder


from uvicore.http.routing.guard import Guard
#from uvicore.auth.middleware.auth import Guard

# from fastapi import APIRouter as _FastAPIRouter
# from uvicore.typing import Any, Type, List, Callable, Optional
# from starlette.routing import BaseRoute
# from uvicore.contracts import ApiRouter as RouterInterface
# from starlette.responses import Response
# from uvicore.support.dumper import dump, dd
# from uvicore.support.module import load


@uvicore.service()
class ApiRouter(Router['ApiRoute']):

    def get(self,
        # Common to both ApiRouter and WebRouter
        path: str,
        endpoint: Optional[Callable] = None,
        *,
        name: Optional[str] = None,
        autoprefix: bool = True,
        middleware: Optional[List] = None,
        auth: Optional[Guard] = None,
        scopes: Optional[List] = None,
        inherits: Optional[Callable] = None,

        # ApiRouter specific
        responses: Optional[Dict] = None,
        response_model: Optional[Any] = None,
        response_class: Optional[Type[Response]] = None,
        tags: Optional[List[str]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
    ) -> None:
        """Add a new HTTP GET route to the ApiRouter

        :param path: URL path, beginning with /
        :param endpoint: None if decorator, else method to call when route is fired.
        :param name: Name of this route to consistently access with url('name') helper.
        :param autoprefix: Disable the name autoprefixer (autoprefix adds the appname. prefix).  Useful when overriding another packages route.
        :param middleware: List of route level middleware.
        :param auth: Shorthand for the middleware=[Guard(['scope'])].  Usage auth=Guard(['scope']).
        :param scopes: Shorthand for middleware=[Guard(['scope'])].  Usage scopes=['scope1', 'scope1'].
        :param response_model: Response ORM Model.  Can also use -> Model on function return type.
        :param tags: List of tags to group the endpoint in the OpenAPI docs.
        :param summary: Summary of this endpoint for OpenAPI docs.
        :param description: Description of this endpoint for OpenAPI docs.
        :param inherits: Endpoint function can inhert the parameters of another function.  Useful if you have tons of parameters you want shared on multiple endpoints. Usage inherits=AuthApi.getsig.
        """
        # Build parameters
        methods = ['GET']
        #params = {key:value for key, value in locals().items() if key != 'self'}
        params = locals()
        params.pop('self')

        # Pass to generic add method
        return self.add(**params)

    def post(self,
        # Common to both ApiRouter and WebRouter
        path: str,
        endpoint: Optional[Callable] = None,
        *,
        name: Optional[str] = None,
        autoprefix: bool = True,
        middleware: Optional[List] = None,
        auth: Optional[Guard] = None,
        scopes: Optional[List] = None,
        inherits: Optional[Callable] = None,

        # ApiRouter specific
        responses: Optional[Dict] = None,
        response_model: Optional[Any] = None,
        response_class: Optional[Type[Response]] = None,
        tags: Optional[List[str]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
    ):
        """Add a new HTTP POST route to the ApiRouter

        :param path: URL path, beginning with /
        :param endpoint: None if decorator, else method to call when route is fired.
        :param name: Name of this route to consistently access with url('name') helper.
        :param autoprefix: Disable the name autoprefixer (autoprefix adds the appname. prefix).  Useful when overriding another packages route.
        :param middleware: List of route level middleware.
        :param auth: Shorthand for the middleware=[Guard(['scope'])].  Usage auth=Guard(['scope']).
        :param scopes: Shorthand for middleware=[Guard(['scope'])].  Usage scopes=['scope1', 'scope1'].
        :param response_model: Response ORM Model.  Can also use -> Model on function return type.
        :param tags: List of tags to group the endpoint in the OpenAPI docs.
        :param summary: Summary of this endpoint for OpenAPI docs.
        :param description: Description of this endpoint for OpenAPI docs.
        :param inherits: Endpoint function can inhert the parameters of another function.  Useful if you have tons of parameters you want shared on multiple endpoints. Usage inherits=AuthApi.getsig.
        """
        # Build parameters
        methods = ['POST']
        #params = {key:value for key, value in locals().items() if key != 'self'}
        params = locals()
        params.pop('self')

        # Pass to generic add method
        return self.add(**params)

    def put(self,
        # Common to both ApiRouter and WebRouter
        path: str,
        endpoint: Optional[Callable] = None,
        *,
        name: Optional[str] = None,
        autoprefix: bool = True,
        middleware: Optional[List] = None,
        auth: Optional[Guard] = None,
        scopes: Optional[List] = None,
        inherits: Optional[Callable] = None,

        # ApiRouter specific
        responses: Optional[Dict] = None,
        response_model: Optional[Any] = None,
        response_class: Optional[Type[Response]] = None,
        tags: Optional[List[str]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
    ):
        """Add a new HTTP PUT route to the ApiRouter

        :param path: URL path, beginning with /
        :param endpoint: None if decorator, else method to call when route is fired.
        :param name: Name of this route to consistently access with url('name') helper.
        :param autoprefix: Disable the name autoprefixer (autoprefix adds the appname. prefix).  Useful when overriding another packages route.
        :param middleware: List of route level middleware.
        :param auth: Shorthand for the middleware=[Guard(['scope'])].  Usage auth=Guard(['scope']).
        :param scopes: Shorthand for middleware=[Guard(['scope'])].  Usage scopes=['scope1', 'scope1'].
        :param response_model: Response ORM Model.  Can also use -> Model on function return type.
        :param tags: List of tags to group the endpoint in the OpenAPI docs.
        :param summary: Summary of this endpoint for OpenAPI docs.
        :param description: Description of this endpoint for OpenAPI docs.
        :param inherits: Endpoint function can inhert the parameters of another function.  Useful if you have tons of parameters you want shared on multiple endpoints. Usage inherits=AuthApi.getsig.
        """
        # Build parameters
        methods = ['PUT']
        #params = {key:value for key, value in locals().items() if key != 'self'}
        params = locals()
        params.pop('self')

        # Pass to generic add method
        return self.add(**params)

    def patch(self,
        # Common to both ApiRouter and WebRouter
        path: str,
        endpoint: Optional[Callable] = None,
        *,
        name: Optional[str] = None,
        autoprefix: bool = True,
        middleware: Optional[List] = None,
        auth: Optional[Guard] = None,
        scopes: Optional[List] = None,
        inherits: Optional[Callable] = None,

        # ApiRouter specific
        responses: Optional[Dict] = None,
        response_model: Optional[Any] = None,
        response_class: Optional[Type[Response]] = None,
        tags: Optional[List[str]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
    ):
        """Add a new HTTP PATCH route to the ApiRouter

        :param path: URL path, beginning with /
        :param endpoint: None if decorator, else method to call when route is fired.
        :param name: Name of this route to consistently access with url('name') helper.
        :param autoprefix: Disable the name autoprefixer (autoprefix adds the appname. prefix).  Useful when overriding another packages route.
        :param middleware: List of route level middleware.
        :param auth: Shorthand for the middleware=[Guard(['scope'])].  Usage auth=Guard(['scope']).
        :param scopes: Shorthand for middleware=[Guard(['scope'])].  Usage scopes=['scope1', 'scope1'].
        :param response_model: Response ORM Model.  Can also use -> Model on function return type.
        :param tags: List of tags to group the endpoint in the OpenAPI docs.
        :param summary: Summary of this endpoint for OpenAPI docs.
        :param description: Description of this endpoint for OpenAPI docs.
        :param inherits: Endpoint function can inhert the parameters of another function.  Useful if you have tons of parameters you want shared on multiple endpoints. Usage inherits=AuthApi.getsig.
        """
        # Build parameters
        methods = ['PATCH']
        #params = {key:value for key, value in locals().items() if key != 'self'}
        params = locals()
        params.pop('self')

        # Pass to generic add method
        return self.add(**params)

    def delete(self,
        # Common to both ApiRouter and WebRouter
        path: str,
        endpoint: Optional[Callable] = None,
        *,
        name: Optional[str] = None,
        autoprefix: bool = True,
        middleware: Optional[List] = None,
        auth: Optional[Guard] = None,
        scopes: Optional[List] = None,
        inherits: Optional[Callable] = None,

        # ApiRouter specific
        responses: Optional[Dict] = None,
        response_model: Optional[Any] = None,
        response_class: Optional[Type[Response]] = None,
        tags: Optional[List[str]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
    ):
        """Add a new HTTP DELETE route to the ApiRouter

        :param path: URL path, beginning with /
        :param endpoint: None if decorator, else method to call when route is fired.
        :param name: Name of this route to consistently access with url('name') helper.
        :param autoprefix: Disable the name autoprefixer (autoprefix adds the appname. prefix).  Useful when overriding another packages route.
        :param middleware: List of route level middleware.
        :param auth: Shorthand for the middleware=[Guard(['scope'])].  Usage auth=Guard(['scope']).
        :param scopes: Shorthand for middleware=[Guard(['scope'])].  Usage scopes=['scope1', 'scope1'].
        :param response_model: Response ORM Model.  Can also use -> Model on function return type.
        :param tags: List of tags to group the endpoint in the OpenAPI docs.
        :param summary: Summary of this endpoint for OpenAPI docs.
        :param description: Description of this endpoint for OpenAPI docs.
        :param inherits: Endpoint function can inhert the parameters of another function.  Useful if you have tons of parameters you want shared on multiple endpoints. Usage inherits=AuthApi.getsig.
        """
        # Build parameters
        methods = ['DELETE']
        #params = {key:value for key, value in locals().items() if key != 'self'}
        params = locals()
        params.pop('self')

        # Pass to generic add method
        return self.add(**params)

    def add(self,
        # Common to both ApiRouter and WebRouter
        path: str,
        endpoint: Optional[Callable] = None,
        methods: List[str] = ['GET'],
        *,
        name: Optional[str] = None,
        autoprefix: bool = True,
        middleware: Optional[List] = None,
        auth: Optional[Guard] = None,
        scopes: Optional[List] = None,
        inherits: Optional[Callable] = None,

        # ApiRouter specific
        responses: Optional[Dict] = None,
        response_model: Optional[Any] = None,
        response_class: Optional[Type[Response]] = None,
        tags: Optional[List[str]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
    ) -> Callable[[Decorator], Decorator]:
        """Add a new HTTP method route to the ApiRouter

        :param path: URL path, beginning with /
        :param endpoint: None if decorator, else method to call when route is fired.
        :param name: Name of this route to consistently access with url('name') helper.
        :param autoprefix: Disable the name autoprefixer (autoprefix adds the appname. prefix).  Useful when overriding another packages route.
        :param middleware: List of route level middleware.
        :param auth: Shorthand for the middleware=[Guard(['scope'])].  Usage auth=Guard(['scope']).
        :param scopes: Shorthand for middleware=[Guard(['scope'])].  Usage scopes=['scope1', 'scope1'].
        :param response_model: Response ORM Model.  Can also use -> Model on function return type.
        :param tags: List of tags to group the endpoint in the OpenAPI docs.
        :param summary: Summary of this endpoint for OpenAPI docs.
        :param description: Description of this endpoint for OpenAPI docs.
        :param inherits: Endpoint function can inhert the parameters of another function.  Useful if you have tons of parameters you want shared on multiple endpoints. Usage inherits=AuthApi.getsig.
        """
        # Convert auth and scope helper param to middleware
        if middleware is None: middleware = []
        if auth: middleware.append(auth)
        if scopes: middleware.append(Guard(scopes))

        # Clean path and name
        (path, full_path, name, full_name) = self._format_path_name(path, name, autoprefix, methods)

        def handle(endpoint):
            nonlocal response_model
            nonlocal description

            # Merge function arguments if inheriting
            if inherits: endpoint = merge_args(inherits, endpoint)

            # If endpoint is partial, grab inside func for type hint and docstrings
            endpoint_func = endpoint
            if isinstance(endpoint, partial):
                # Endpoint is a partial (was overwritten to default some higher order middleware)
                # A partial overwrites the original docstring.  Functools update_wrapper will copy it back
                # as well as handle merging of other important properties
                #update_wrapper(route.endpoint, route.endpoint.func)
                endpoint_func = endpoint.func

                # Blank out the __doc__ on actual Partial itself, not actual endpoint inside partial.
                # If not, OpenAPI doc description will be partial(func, *args, **keywords) - new function with partial application etc...
                endpoint.__doc__ = None

            # Get response_model from typehint if not explicitly defined as a parameter
            response_model = response_model or get_type_hints(endpoint_func).get('return')

            # Get openapi description from route param or endpoint docstring
            description = description or endpoint_func.__doc__

            # Create route SuperDict
            route = ApiRoute({
                # Common to both ApiRouter and WebRouter
                'path': full_path,
                'name': full_name,
                'methods': methods,
                'endpoint': endpoint,
                'middleware': middleware,
                'autoprefix': autoprefix,
                'original_path': path,
                'original_name': name,
                'responses': responses,

                # ApiRouter specific
                'response_model': response_model,
                'response_class': response_class,
                'tags': tags,
                'summary': summary,
                'description': description,
            })
            # Full name already contains -POST, -PUT...for uniqueness across methods
            self.routes[full_name] = route
            return route

        # Method access
        if endpoint: return handle(endpoint)

        # Decorator access
        def decorator(func: Decorator) -> Decorator:
            handle(func)
            return func
        return decorator

@uvicore.service()
class ApiRoute(RouteInterface):
    """Route superdict definition"""
    pass


@register_pretty(ApiRoute)
def pretty_entity(value, ctx):
    """Custom pretty printer for my SuperDict"""
    # This printer removes the class name uvicore.types.Dict and makes it print
    # with a regular {.  This really cleans up the output!

    # SuperDict are printed as Dict, but this Package SuperDict should
    # be printed more like a class with key=value notation, so use **values
    return pretty_call(ctx, 'ApiRoute', **value)


# # Example of just using the entier FastAPI Router without abstraction
# @uvicore.service()
# class ApiRouter(_FastAPIRouter):

#     def resource(self, module, *, prefix: str = '', tags:List[str] = None) -> None:
#         """Include Controller Routes"""
#         if type(module) == str:
#             # Using a string to point to an endpoint class controller
#             module_path = module
#             if self.uvicore.endpoints: module_path = self.uvicore.endpoints + '.' + module_path
#             controller = load(module_path)
#             uvicore.app.http.include_router(
#                 controller.object,
#                 prefix=self.uvicore.prefix + str(prefix),
#                 tags=tags,
#             )
#         else:
#             # Passing in an actual router class
#             uvicore.app.http.include_router(
#                 module,
#                 prefix=self.uvicore.prefix + str(prefix),
#                 tags=tags,
#             )



# # @route.get('/about', 'about')
# # async def about(request: Request):
# #     # Example Jinja2 Template
# #     return response.View('wiki/about.j2', {'request': request})


# # Example of making my own router abstraction
# class ApiRouterX(RouterInterface):

#     @property
#     def router(self) -> _FastAPIRouter:
#         return self._router

#     # These properties have to exist because the FastAPI code looks for route.routes
#     # or routes.response_class for example so this router abstraction must pass them through
#     # Perhaps my own abstraction isn't a good idea?  Maybe just pass through fastAPI's router?
#     @property
#     def routes(self) -> List[BaseRoute]:
#         return self._router.routes

#     @property
#     def response_class(self) -> Type[Response]:
#         return self._router.response_class

#     @property
#     def default_response_class(self) -> Type[Response]:
#         return self._router.default_response_class

#     @property
#     def on_startup(self) -> None:
#         return self._router.on_startup

#     @property
#     def on_shutdown(self) -> None:
#         return self._router.on_shutdown

#     def __init__(self):
#         # Fireup FastAPI Router
#         self._router = _FastAPIRouter()

#     def get(self,
#         path: str,
#         name: str = None,
#         *,
#         response_model: Type[Any] = None,
#         include_in_schema: bool = True
#     ) -> Callable:
#         return self._router.get(
#             path=path,
#             name=name,
#             response_model=response_model,
#             include_in_schema=include_in_schema,
#         )

#     #def include_router(self, router: "APIRouter", *, prefix: str = '', tags: List[str] = None) -> None:
#     # def include_router(self, router: "APIRouter") -> None:
#     #     # Only used for the class based controller
#     #     # return self._router.include_router(router,
#     #     #     prefix=prefix,
#     #     #     tags=tags
#     #     # )
#     #     return self.router.include_router(router)

#     def add_api_route(self, path: str, endpoint: Callable, *,
#         response_model: Optional[Type[Any]] = None,
#         name: Optional[str] = None,
#         # self,
#         # path: str,
#         # endpoint: Callable,
#         # *,
#         # response_model: Optional[Type[Any]] = None,
#         # status_code: int = 200,
#         # tags: Optional[List[str]] = None,
#         # dependencies: Optional[Sequence[params.Depends]] = None,
#         # summary: Optional[str] = None,
#         # description: Optional[str] = None,
#         # response_description: str = "Successful Response",
#         # responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
#         # deprecated: Optional[bool] = None,
#         # methods: Optional[Union[Set[str], List[str]]] = None,
#         # operation_id: Optional[str] = None,
#         # response_model_include: Optional[Union[SetIntStr, DictIntStrAny]] = None,
#         # response_model_exclude: Optional[Union[SetIntStr, DictIntStrAny]] = None,
#         # response_model_by_alias: bool = True,
#         # response_model_exclude_unset: bool = False,
#         # response_model_exclude_defaults: bool = False,
#         # response_model_exclude_none: bool = False,
#         # include_in_schema: bool = True,
#         # response_class: Optional[Type[Response]] = None,
#         # name: Optional[str] = None,
#         # route_class_override: Optional[Type[APIRoute]] = None,
#         # callbacks: Optional[List[APIRoute]] = None,
#     ) -> None:
#         return self.router.add_api_route(
#             path=path,
#             endpoint=endpoint,
#             response_model=response_model,
#             name=name,
#             # status_code=status_code,
#             # tags=tags,
#             # dependencies=dependencies,
#             # summary=summary,
#             # description=description,
#             # response_description=response_description,
#             # responses=responses,
#             # deprecated=deprecated,
#             # methods=methods,
#             # operation_id=operation_id,
#             # response_model_include=response_model_include,
#             # response_model_exclude=response_model_exclude,
#             # response_model_by_alias=response_model_by_alias,
#             # response
#         )

#     # @property
#     # def router(self):
#     #     return self._router



# # IoC Class Instance
# #ApiRouter: RouterInterface = uvicore.ioc.make('ApiRouter', _ApiRouter)

# # Public API for import * and doc gens
# #__all__ = ['ApiRouter', '_ApiRouter']



import uvicore
from uvicore.http.request import Request


@uvicore.composer()
class xx_ComposerName:
    """xx_ComposerName view composer"""

    def __init__(self,
        request: Request,
        name: str,
        context: dict,
        status_code: int,
        headers: dict,
        media_type: str
    ) -> None:
        self.request = request
        self.name = name
        self.context = context
        self.status_code = status_code
        self.headers = headers
        self.media_type = media_type

    async def compose(self) -> dict:
        # This composer is registered in your service provider boot() method.
        # Example: self.composer('xx_vendor.xx_appname.http.composers.xx_composername.xx_ComposerName', ['xx_appname/*'])

        # Return new context to merge into views
        return {
            'example': 'available to your views that match the composer wildcard',
        }



from __future__ import annotations
import uvicore
from uvicore.typing import Optional, List
from uvicore.auth.models.role import Role
from uvicore.support.dumper import dd, dump
from uvicore.orm import Model, ModelMetaclass, Field, BelongsToMany
from uvicore.auth.database.tables import groups as table


@uvicore.model()
class Group(Model['Group'], metaclass=ModelMetaclass):
    """Auth Group Model"""

    __tableclass__ = table.Groups

    id: Optional[int] = Field('id',
        primary=True,
        description='Group ID',
    )

    # key: str = Field('key',
    #     primary=True,
    #     description='Group Primary Key',
    # )

    name: str = Field('name',
        description='Group Name',
    )

    # Many-To-Many via group_roles pivot table
    roles: Optional[List[Role]] = Field(None,
        description="Group Roles",
        relation=BelongsToMany('uvicore.auth.models.role.Role', join_tablename='group_roles', left_key='group_id', right_key='role_id'),
    )



import uvicore
from uvicore.typing import Dict, List, OrderedDict, get_type_hints, Tuple
from uvicore.events import Handler
from uvicore.support import module
from uvicore.support.dumper import dump, dd
from uvicore.foundation.events.app import Booted as OnAppBooted
from uvicore.contracts import Package as Package
from uvicore.console import command_is
from starlette.applications import Starlette
from fastapi import FastAPI, Depends
from uvicore.http import response
from uvicore.http.events import server as HttpServerEvents
from uvicore.http.routing.router import Routes
from functools import partial, update_wrapper
from uvicore.http.routing import ApiRoute, WebRoute
from uvicore.http.routing import Guard
from uvicore.http.openapi.docs import get_swagger_ui_html, get_swagger_ui_oauth2_redirect_html


class Templating(Handler):

    def __call__(self, event: OnAppBooted):
        """Bootstrap the Templating System after the Application is Booted"""

        # Initialize Templating System
        self.initialize_templates()


    def initialize_templates(self) -> None:
        """Initialize Templating System"""

        # Loop each package with an HTTP definition and add to our HTTP server
        paths = []
        context_processors = Dict()
        for package in uvicore.app.packages.values():
            if not 'templating' in package: continue

            # Append template paths
            for path in package.templating.paths or []:
                paths.append(path)

            # Deep merge template context_processors
            context_processors.merge(package.templating.context_processors or {})

        # Get the template singleton from the IoC
        templates = uvicore.ioc.make('uvicore.templating.engine.Templates')

        # Add all packages view paths
        # First path wins, so we must REVERSE the package order
        # This will set the main running app FIRST as we always want the app to win
        paths.reverse()
        for path in paths:
            templates.include_path(module.location(path))

        # Add all packages deep_merged template options
        if 'context_functions' in context_processors:
            for name, method in context_processors['context_functions'].items():
                templates.include_context_function(name, method)
        if 'context_filters' in context_processors:
            for name, method in context_processors['context_filters'].items():
                templates.include_context_filter(name, method)
        if 'filters' in context_processors:
            for name, method in context_processors['filters'].items():
                templates.include_filter(name, method)
        if 'tests' in context_processors:
            for name, method in context_processors['tests'].items():
                templates.include_test(name, method)

        # Initialize template system
        templates._init()



import uvicore
from uvicore.http import Request, response
from uvicore.http.routing import WebRouter, Controller

# Extra
# from uvicore.auth import UserInfo
# from uvicore.http.routing import Guard
# from uvicore.typing import Dict, List, Optional
# from uvicore.http.exceptions import HTTPException
# from uvicore.http.params import Path, Query, Header, Cookie, Body, Form, File, Depends, Security


# ------------------------------------------------------------------------------
# Uvicore Controller Schematic
# This schematic is filled with examples, a suppliment to the docs.
# Pick the best example for your use case and modify as needed!
# ------------------------------------------------------------------------------


@uvicore.controller()
class xx_ControllerName(Controller):

    # --------------------------------------------------------------------------
    # Example:  Multiple ways to apply route level middleware including
    # auth guards with scoped permissions to this entire controller.
    # Tip: Your routes/api.py and routes/web.py are actually the same
    # as controllers.  In fact controllers are just nested routers.  So all
    # notes here also apply to your routes/* files and controllers.
    # --------------------------------------------------------------------------
    # Apply scopes to all routes and children controllers - simple and preferred
    #scopes = ['authenticated', 'employee']

    # Scopes is a shortcut to
    #auth = Guard(['authenticated'])

    # Scopes and auth are shortcuts to the full middleware stack
    # middleware = [
    #     Guard(['authenticated']),
    #     Any other route based middleware here
    # ]

    # Optionally, apply scopes and also grab the current user
    # user: UserInfo = Guard(['authenticated'])
    #   Then in your routes, inject the user with
    #   async def welcome(request: Request, user: UserInfo = self.user):

    def register(self, route: WebRouter):
        """Register Web Controller Endpoints"""

        # ----------------------------------------------------------------------
        # Example: Basic route responding with a view template
        # ----------------------------------------------------------------------
        # @route.get('/example1/{id}')
        # async def example1(request: Request, id: int):
        #     # Template should be in http/views/xx_appname/template.j2
        #     # All web routes require request: Request
        #     # All template views require the request be piped through
        #     # To see all route details run ./uvicore package list
        #     return await response.View('xx_appname/template.j2', {
        #         'request': request,
        #         'id': id,
        #     })


        # ----------------------------------------------------------------------
        # Example: Other types of responses
        # ----------------------------------------------------------------------
        # @route.get('/example2')
        # async def example2(request: Request):
        #     return response.Text('Text Here')
        #     return response.HTML('<b>HTML</b> here')
        #     return response.JSON({'json':'here'})
        #     return response.UJSON({'json':'here'}) # requires ujson dependency
        #     return response.ORJSON({'json':'here'}) # requires orjson dependency
        #     # and more ... see uvicore/http/response.py


        # ----------------------------------------------------------------------
        # Example: GET request variables
        # curl http://localhost/?name=asdf
        # ----------------------------------------------------------------------
        # @route.get('/example2')
        # async def example2(request: Request, name: str):
        #     return response.Text(name)


        # ----------------------------------------------------------------------
        # Example: Changing the route name
        # Each route is given a name automatically.  A name is what you can use
        # to reference the route in your code and view templates.  Try to use the
        # route name instead of the path as paths WILL change as other users use
        # your library because they can tweak the BASE PATH.  Views should be using
        # {{ url('xx_appname.ex2a')}}, never /example2a
        # If you don't specify a name, uvicore makes a name automatically
        # from the path, even nested paths from groups.  Name always starts
        # with your apps name, ie: xx_appname.
        # ----------------------------------------------------------------------
        # @route.get('/example2a', name='ex2a')
        # async def example2a(request: Request):
        #     # Route name is xx_appname.ex2a
        #     return response.Text('example2a')


        # ----------------------------------------------------------------------
        # Example: Override the naming autoprefix
        # When you define a custom route name, a prefix of xx_appname. is
        # automatically added keeping your route scoped to this package.
        # In order to set a full name you must set autoprefix=False.
        # This is handy when you want to override a route from another package.
        # ----------------------------------------------------------------------
        # @route.get('/example2b', name='someother.app.ex2b', autoprefix=False)
        # async def example2b(request: Request):
        #     # Route name is someother.app.ex2b
        #     return response.Text('example2b')


        # ----------------------------------------------------------------------
        # Example: Injecting the current User using the Guard dependency
        # ----------------------------------------------------------------------
        # @route.get('/example3')
        # async def example3(request: Request, user: UserInfo = Guard()):
        #     dump(user)
        #     return response.Text('example3')


        # ----------------------------------------------------------------------
        # Example: Injecting the current User using the Guard dependency
        # And guarding the route based on the users permissions (scopes)
        # Benifit of this vs decorator scopes is you also get the current User
        # injected so you don't have to use request.scope['user]
        # curl -u user:pass http://localhost/example4
        # ----------------------------------------------------------------------
        # @route.get('/example4')
        # async def example4(request: Request, user: UserInfo = Guard(['authenticated'])):
        #     dump(user);
        #     return response.Text('example4')


        # ----------------------------------------------------------------------
        # Example: Getting the user from request.scope['user']
        # And guarding the route from the decorator using the scopes shortcut
        # curl -u user:pass http://localhost/example5
        # ----------------------------------------------------------------------
        # @route.get('/example5', scopes=['authenticated'])
        # async def example5(request: Request):
        #     user: UserInfo = request.scope['user'];
        #     dump(user)
        #     return response.Text('example5')


        # ----------------------------------------------------------------------
        # Example: Using the auth guard instead of scopes shortcut
        # while also getting the current user from request.scope['user']
        # curl -u user:pass http://localhost/example5a
        # ----------------------------------------------------------------------
        # @route.get('/example5a', auth=Guard(['authenticated']))
        # async def example5a(request: Request):
        #     user: UserInfo = request.scope['user'];
        #     dump(user)
        #     return response.Text('example5a')


        # ----------------------------------------------------------------------
        # Example: Using the per route middleware instead of auth or scopes
        # curl -u user:pass http://localhost/example5b
        # ----------------------------------------------------------------------
        # @route.get('/example5b', middleware=[
        #     Guard(['authenticated'])
        # ])
        # async def example5b(request: Request):
        #     user: UserInfo = request.scope['user'];
        #     dump(user)
        #     return response.Text('example5b')


        # ----------------------------------------------------------------------
        # Example: POST with form body
        # curl -X POST -F 'name=matthew' http://localhost/
        # http -v --form POST http://localhost/ name=matthew
        # ----------------------------------------------------------------------
        # @route.post('/example6')
        # async def example6(request: Request, name: str = Form(...)):
        #     return response.Text(name)


        # ----------------------------------------------------------------------
        # Example raise proper HTTP Exception
        # ----------------------------------------------------------------------
        # @route.get('/example7')
        # async def example7(request: Request):
        #     raise HTTPException(404, 'bad stuff')


        # ----------------------------------------------------------------------
        # Example: Grouping routes for common paths and scopes
        # Routes will be /group1/example8 and /group1/subgroup1/example9
        # with scopped permissions on both routes.
        # IF you also set name='g1', all route names will be
        # xx_appname.g1.example8 instead of autonamed xx_appname.group1.example8
        # ----------------------------------------------------------------------
        # @route.group('/group1', scopes=['authenticated'])
        # def group1():
        #     @route.get('/example8')
        #     async def example8(request: Request):
        #         return response.Text("example8")

        #     @route.group('/subgroup1')
        #     def subgroup1():
        #         @route.get('/example9')
        #         async def example9(request: Request):
        #             return response.Text("example9")


        # ----------------------------------------------------------------------
        # Example: Routes as method callbacks (no decorators)
        # ----------------------------------------------------------------------
        # def example10(request: Request):
        #     return response.Text('example10')
        # route.get('/example10', example10)


        # ----------------------------------------------------------------------
        # Example: Groups and routes as method callbacks (no decorators)
        # ----------------------------------------------------------------------
        # def example11(request: Request):
        #     return response.Text('example11')
        # def example12(request: Request):
        #     return response.Text('example12')
        # route.group('/group2', scopes=['authenticated'], routes=[
        #     route.get('/example11', example11),
        #     route.get('/example12', example12),
        # ])


        # ----------------------------------------------------------------------
        # Example: Including other route files and controllers
        # Technically, there is no difference between your routes files
        # in http/routes/web.py and http/routes/api.py and your controllers
        # or api controllers.  They are all just one huge nested router.
        # They are only split up for logical convenience.  Just as you included
        # controlers from your http/routes/web.py, you can also include other
        # routes here.
        # ----------------------------------------------------------------------
        # route.controller('xx_vendor.xx_appname.http.controllers.some.other.Other')

        # Also, route.controller and route.include are aliases of each other, same thing.
        # route.include('xx_vendor.xx_appname.http.controllers.some.other2.Other2')

        # Instead of typing the full module path, if route.controllers is defined
        # Then all .controller() and .include() can use relative paths
        # route.controllers = 'xx_vendor.xx_appname.http.controllers'

        # Looks for Class in xx_vendor.xx_appname.http.controllers.some.Some
        # route.controller('some')

        # Leading period means APPEND path to defined route.controllers
        # So this looks for xx_vendor.xx_appname.http.controllers.other3.Other3
        # route.controller('.other3.Other3')

        # If no leading . but other . exists, then it is assuming a full path,
        # regardless if route.controllers is defined or not.


        # Return router
        # Must always return the router at the end of every controller and routes file
        # as this is one infinitely recursive nested router configuration.
        return route



# Notice, this uvicore.http.params is a naming convenience only
# All of these are also available at uvicore.http.request
from fastapi.datastructures import UploadFile, DefaultPlaceholder

# Difference between fastapi.params vs fastapi.param_functions??
# param_functions add more props and defaults then instantiate
# the param Class themselves.  Like a higher wrapper around the classes.
# All docs say 'from fastapi import Form' which actually comes from param_functions
from fastapi.param_functions import Path
from fastapi.param_functions import Query
from fastapi.param_functions import Header
from fastapi.param_functions import Cookie
from fastapi.param_functions import Body
from fastapi.param_functions import Form
from fastapi.param_functions import File
from fastapi.param_functions import Depends
from fastapi.param_functions import Security



from __future__ import annotations
from abc import ABC, abstractmethod, abstractproperty, abstractclassmethod
from uvicore.typing import Dict, List, Union
from prettyprinter import pretty_call, register_pretty
from dataclasses import dataclass

# Why a dataclass instead of a SuperDict?
# Because a dataclass FORCES you to instantiate with ALL these attributes
# making them required.  If one extends the UserInfo all of these are guarenteed
# required in order to instantiate.  This is inherently a good contract.



@dataclass
class UserInfo(ABC):
    """Auth Logged in User Definition"""

    # These class level properties for for type annotations only.
    # They do not restrict of define valid properties like a dataclass would.
    # This is still a fully dynamic SuperDict!
    id: int
    uuid: str
    sub: str
    username: str
    email: str
    first_name: str
    last_name: str
    title: str
    avatar: str
    groups: List[str]
    roles: List[str]
    permissions: List[str]
    superadmin: bool
    authenticated: bool

    @abstractproperty
    def name(self):
        """First and last name"""

    @abstractproperty
    def avatar_url(self):
        """Alias to avatar"""

    @abstractproperty
    def admin(self):
        """Alias to superadmin"""

    @abstractproperty
    def is_admin(self):
        """Alias to superadmin"""

    @abstractproperty
    def is_superadmin(self):
        """Alias to superadmin"""

    @abstractproperty
    def is_authenticated(self):
        """Alias to authenticated"""

    @abstractproperty
    def loggedin(self):
        """Alias to authenticated"""

    @abstractproperty
    def is_loggedin(self):
        """Alias to authenticated"""

    @abstractproperty
    def check(self):
        """Alias to authenticated"""

    def can(self, permissions: Union[str, List]) -> bool:
        """Check if user has ALL of these permissions"""

    def can_any(self, permissions: Union[str, List]) -> bool:
        """Check if user has any one of these permissions"""

    def cant(self, permissions: Union[str, List]) -> bool:
        """Check if user does not have one of these permissions"""

    def cannot(self, permissions: Union[str, List]) -> bool:
        """Alias to cant"""




import uvicore
import sqlalchemy as sa
from uvicore.database import Table
from uvicore.support.dumper import dump, dd


# Get related table names with proper prefixes
# users = uvicore.db.tablename('auth.users')
# formats = uvicore.db.tablename('xx_appname.formats')


@uvicore.table()
class xx_TableName(Table):

    # Actual database table name
    # Plural table names and singluar model names are encouraged
    # Do not add a package prefix, leave that to the connection config
    name = 'xx_tablename'

    # Connection for this database from your config file
    connection = 'xx_appname'

    # SQLAlchemy Table definition as a list (exclude name and metadata)
    # This will be converted into an actual SQLAlchemy Table() instance
    # See https://docs.sqlalchemy.org/en/13/core/schema.html
    schema = [
        # Defaults: nullable=True, index=False, unique=False, primary_key=False

        sa.Column('id', sa.Integer, primary_key=True),  # SQL auto incrementing primary key int(11) datatype, the "clustered"index
        sa.Column('slug', sa.String(length=100), unique=True),  # SQL varchar(100) with unique index
        sa.Column('title', sa.String(length=100)),  # SQL varchar(100) datatype

        # sa.Column('key', sa.CHAR(length=3), primary_key=True),  # SQL non-incrementing primary key char(3) datatype, the "clustered"index

        # sa.Column('body', sa.Text()),  # SQL text datatype

        # sa.Column('view_count', sa.Integer),  # SQL int(11) datatype
        # sa.Column('view_count', sa.CHAR(length=3)),  # SQL char(4) datatype
        # sa.Column('dollar_amt', sa.DECIMAL(precision=8, scale=2)),  # SQL decimal(8,2) datatype max 999,999.99
        # sa.Column('dollar_amt2', sa.Float()),  # SQL Float datatype

        # sa.Column('is_deleted', sa.Boolean(), default=False),  # SQL tinyint(1) datatype

        # sa.Column('created_at', sa.DateTime()), # SQL datetime datatype
        # sa.Column('updated_at', sa.DateTime()), # SQL datetime datatype

        # sa.Column('format', sa.Integer, sa.ForeignKey(f"{formats}.id"), nullable=False),   # SQL Foreign key, which does create an index
        # sa.Column('creator_id', sa.Integer, sa.ForeignKey(f"{users}.id"), nullable=False), # SQL Foreign key, which does create an index
        # sa.Column('updator_id', sa.Integer, sa.ForeignKey(f"{users}.id"), nullable=False), # SQL Foreign key, which does create an index

        # Multi Column Unique Constraint.  By adding in the key we still ensure
        # OneToMany can be used but it must be unique with the key.  This also creates
        # a good composite index of type,id,key
        # sa.UniqueConstraint('attributable_type', 'attributable_id', 'key'),

        # If you don't want an ID primary_key, you could use the combined poly IDs as a PK
        # But the ORM can't handle duel PKs at the moment
        # sa.PrimaryKeyConstraint('attributable_type', 'attributable_id', 'key'),
    ]

    # Optional SQLAlchemy Table() instance kwargs
    schema_kwargs = {}



import uvicore
from uvicore.events import Event


@uvicore.event()
class Startup(Event):
    """Console is starting up.  Runs before console command."""

    is_async = True

    def __init__(self):
        pass


@uvicore.event()
class PytestStartup(Event):
    """Pytest console is starting up.  Runs before pytests."""

    is_async = True

    def __init__(self):
        pass


@uvicore.event()
class Shutdown(Event):
    """Console is shutting down.  Runs after console command"""

    is_async = True

    def __init__(self):
        pass


@uvicore.event()
class PytestShutdown(Event):
    """Pytest console is shutting down.  Runs after pytests"""

    is_async = True

    def __init__(self):
        pass




import uvicore
from typing import Any
from uvicore.support.dumper import dump, dd


def url(context: dict, name: str, **path_params: Any) -> str:
    # Get http configuration
    config = uvicore.config.app.http

    # Get request from context
    request = context["request"]

    # Get full URL from starlette's request url_for() method
    try:
        url = request.url_for(name, **path_params)
    except Exception as e:
        raise Exception('Could not find URL for {} in template {}'.format(name, context.name))

    # Ensure url contains proper x-forwarded-proto protocol
    # If serving with uvicorn this is detected properly, but if serving with
    # gunicorn it is not.  So we'll double check and force HTTPS if needed.
    if 'x-forwarded-proto' in request.headers:
        proto = request.headers['x-forwarded-proto'].lower()
        if proto == 'https' and proto not in url:
            # Starlette generated URL does not contain https, but it should
            url = 'https' + url[4:]

    # Return full URL
    return url

def asset(context: dict, path: str) -> str:

    # Detect custom host and path
    asset_host = uvicore.config.app.web.asset.host
    asset_path = uvicore.config.app.web.asset.path or '/assets'
    asset_force_https = uvicore.config.app.web.asset.force_https or False

    if asset_host:
        # Using a custom asset host
        return asset_host + asset_path + '/' + path
    else:
        # Using domain of running app
        return url(context, 'assets', path=path)

def public(context: dict, path: str) -> str:
    return url(context, 'public', path=path)




import sqlalchemy as sa

import uvicore
from uvicore.database import Table
from uvicore.support.dumper import dd, dump


# Get related tablenames with proper prefixes
users = uvicore.db.tablename('auth.users')

@uvicore.table()
class Users(Table):

    # Actual database table name
    # Plural table names and singluar model names are encouraged
    # Do not add a package prefix, leave that to the connection config
    name = 'users'

    # Connection for this database from your config file
    connection = 'auth'

    # SQLAlchemy Table definition as a list (exclude name and metadata)
    # This will be converted into an actual SQLAlchemy Table() instance
    # See https://docs.sqlalchemy.org/en/13/core/schema.html
    schema = [
        sa.Column('id', sa.Integer, primary_key=True),
        sa.Column('uuid', sa.String(length=36), unique=True),
        sa.Column('username', sa.String(length=50), unique=True),
        sa.Column('email', sa.String(length=50), unique=True),
        sa.Column('first_name', sa.String(length=30)),
        sa.Column('last_name', sa.String(length=30)),
        sa.Column('title', sa.String(length=50)),
        sa.Column('avatar_url', sa.String(length=500)),
        sa.Column('password', sa.Text()),
        sa.Column('disabled', sa.Boolean, default=False),
        sa.Column('creator_id', sa.Integer, sa.ForeignKey(f"{users}.id")),
        sa.Column('created_at', sa.DateTime, server_default=sa.func.now()),
        sa.Column('updated_at', sa.DateTime, server_default=sa.func.now(), onupdate=sa.func.now()),
        sa.Column('login_at', sa.DateTime, nullable=True),


        # 10 dddddddddd
        # 20 dddddddddddddddddddd
        # 30 dddddddddddddddddddddddddddddd
        # 40 dddddddddddddddddddddddddddddddddddddddd
        # 50 dddddddddddddddddddddddddddddddddddddddd
        # 100 dddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddd
        # 200 dddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddd

    ]

    # Optional SQLAlchemy Table() instance kwargs
    schema_kwargs = {}



import pickle
import uvicore
from uvicore.typing import Dict, Any, Callable, Union, List, Tuple
from uvicore.support.dumper import dump, dd
from uvicore.redis import Redis as RedisDb
from aioredis import Redis as RedisInterface
from uvicore.contracts import Cache as CacheInterface
from uvicore.cache.manager import Manager


@uvicore.service()
class Redis(CacheInterface):

    def __init__(self, manager: Manager, store: Dict):
        self.manager = manager
        self.connection = store.connection
        self.prefix = store.prefix
        self.seconds = store.seconds
        self._redis = None

    def connect(self, store: str = None) -> CacheInterface:
        """Connect to a cache backend store"""
        return self.manager.connect(store)

    def store(self, store: str = None) -> CacheInterface:
        """Alias to connect"""
        return self.connect(store)

    async def has(self, key: str) -> bool:
        """Check if key exists"""
        (redis, key) = await self._prepair(key)
        return bool(await redis.exists(key))

    async def get(self, key: Union[str, List], *, default: Any = None) -> Any:
        """Get one or more key values if exists else return default value"""
        (redis, keys) = await self._prepair(key)
        if type(keys) == list:
            values = Dict()
            for key in keys:
                return_key = key[len(self.prefix):]
                if await self.has(key):
                    # Item exists, get from redis
                    values[return_key] = self._deserialize(await redis.get(key))
                else:
                    # Item does not exist, set default
                    values[return_key] = default
            return values
        else:
            if await self.has(keys):
                # Item exists, get from redis
                return self._deserialize(await redis.get(keys))
            else:
                # Item does not exist, set default
                return default

    async def remember(self, key: Union[str, Dict], callback: Union[Callable, Any] = None, *, seconds: int = None) -> Any:
        """Get a key if exists, if not SET the key to callback value"""
        (redis, keys) = await self._prepair(key)
        if type(key) != dict: keys = {keys:callback}
        value = {}
        for key, callback in keys.items():
            if await self.has(key):
                # Item exists, simply return value
                #value[key] = await redis.get(key)
                value[key] = await self.get(key)
            else:
                # Item does not exist, set value based on callback return
                if callable(callback):
                    value[key] = await callback()
                else:
                    value[key] = callback
                await self.put(key, value[key], seconds=seconds)

        if type(key) != dict:
            # Single key, single return
            return value[key]

        # Dict key, dict return
        return value

    async def put(self, key: Union[str, Dict], value: Any = None, *, seconds: int = None) -> None:
        """Put one or more key/values in cache with optional expire in seconds (0=never expire)"""
        (redis, keys) = await self._prepair(key)
        if seconds is None: seconds = self.seconds
        if type(keys) != dict: keys = {keys:value}
        for (key, value) in keys.items():
            await redis.set(key, self._serialize(value), expire=seconds)

    async def pull(self, key: Union[str, Dict]) -> Any:
        """Get one or more key values from cache them remove them after"""
        (redis, keys) = await self._prepair(key)
        value = await self.get(key)
        await self.forget(keys)
        return value

    async def add(self, key: str, value: Any, *, seconds: int = None) -> bool:
        """Put a single value in cache only if not exists"""
        (redis, key) = await self._prepair(key)
        if await self.has(key):
            # Item already exists, return False for NOT added
            return False
        else:
            # Item does not exist, set value and return True
            await self.put(key, value, seconds=seconds)
            return True

    async def touch(self, key: str, *, seconds: int = None) -> bool:
        """Touch a key, if seconds are provided, also reset expire TTL"""
        (redis, key) = await self._prepair(key)
        if await self.has(key):
            await redis.touch(key)
            if seconds is not None:
                await redis.expire(key, seconds)
            return True
        return False

    async def increment(self, key, by: int = 1, *, seconds: int = None) -> int:
        """Increment a key by integer specified.  If key not exists, sets key to increment value"""
        (redis, key) = await self._prepair(key)
        if seconds is None: seconds = self.seconds
        await redis.incrby(key, by)
        if seconds == 0:
            # 0 means never expire
            await redis.persist(key)
        else:
            await redis.expire(key, seconds)
        return int(await redis.get(key))

    async def decrement(self, key, by: int = 1, *, seconds: int = None) -> int:
        """Decrement a key by integer specified.  If key not exists, sets key to decrement value"""
        (redis, key) = await self._prepair(key)
        if seconds is None: seconds = self.seconds
        await redis.decrby(key, by)
        if seconds == 0:
            # 0 means never expire
            await redis.persist(key)
        else:
            await redis.expire(key, seconds)
        return int(await redis.get(key))

    async def forget(self, key: Union[str, List]) -> None:
        """Delete a key from cache"""
        (redis, keys) = await self._prepair(key)
        if type(keys) != list: keys = [keys]
        await redis.delete(*keys)

    async def flush(self) -> None:
        """Flush entire cache.  Only deletes keys with proper cache prefix."""
        redis = (await self._prepair())[0]
        keys = await redis.keys(self.prefix + '*')
        await redis.delete(*keys)

    async def _prepair(self, key: Union[str, List] = None) -> Tuple[RedisInterface, Union[str, List, Dict]]:
        # Connect to redis pool if not connected
        if not self._redis:
            # RedisDB is uvicore.redis.redis.py abstraction and .connect() there
            # runs the await aioredis.create_redis_pool if not connected
            self._redis = await RedisDb.connect(self.connection)

        if key:
            if type(key) == list:
                # Check if prefix already added
                key0 = str(key[0])
                prefix = self.prefix
                if len(key0) > len(prefix) and key0[0:len(prefix)] == prefix: prefix = ''
                return (self._redis, [prefix + k for k in key])
            elif type(key) == dict:
                key0 = [x for x in key.keys()][0]
                prefix = self.prefix
                if len(key0) > len(prefix) and key0[0:len(prefix)] == prefix: prefix = ''
                return (self._redis, {prefix + k:v for k,v in key.items()})
            else:
                # Check if prefix already added
                key = str(key)
                prefix = self.prefix
                if len(key) > len(prefix) and key[0:len(prefix)] == prefix: prefix = ''
                return (self._redis, prefix + key)
        else:
            return (self._redis, None)

    def _serialize(self, value):
        return pickle.dumps(value)

    def _deserialize(self, value):
        # Error here means value was never serialized.
        # Like with .increment and .decrement keys
        try:
            return pickle.loads(value)
        except:
            return value.decode()




import uvicore
from uvicore.typing import Any
from uvicore.support.dumper import dump, dd
from uvicore.contracts import JobDispatcher as JobDispatcherInterface

@uvicore.service('uvicore.jobs.dispatcher.Dispatcher',
    aliases=['JobDispatcher'],
    singleton=True
)
class Dispatcher(JobDispatcherInterface):

    def dispatch(self, instance: object) -> Any:
        """Dispatch a Job Class"""
        return instance.handle()

    async def dispatch_async(self, instance: object) -> Any:
        return await instance.handle()

    async def codispatch(self, instance: object) -> Any:
        """Alias for dispatch_async()"""
        return await self.dispatch_async(instance)



from typing import Optional

import uvicore
from uvicore import app, log
from uvicore.console import command
from uvicore.support.dumper import dd, dump

@command()
def list(): # pragma: no cover
    """Show all Web and API Routes"""
    log.header("Final Merged Web and API Routes")
    log.line()
    dd({
        'web': app.running_config.http.web.routes,
        'api': app.running_config.http.api.routes,
    })




from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Generic, List, Tuple, TypeVar, Union, Dict

try:
    import sqlalchemy as sa
    from sqlalchemy.engine.result import RowProxy
    from sqlalchemy.sql.expression import BinaryExpression
except ImportError:  # pragma: nocover
    sa = None  # type: ignore
    RowProxy = None  # type: ignore
    BinaryExpression = None  # type: ignore

B = TypeVar("B")  # Builder Type (DbQueryBuilder or OrmQueryBuilder)
E = TypeVar("E")  # Entity Model


class QueryBuilder(Generic[B, E], ABC):
    @abstractmethod
    def where(self, column: Union[str, BinaryExpression, List[Union[Tuple, BinaryExpression]]], operator: str = None, value: Any = None) -> B[B, E]:
        """Add where statement to query"""

    @abstractmethod
    def or_where(self, wheres: List[Union[Tuple, BinaryExpression]]) -> B[B, E]:
        """Add or where statement to query"""

    @abstractmethod
    def order_by(self, column: Union[str, List[str], List[Tuple], Any], order: str = 'ASC') -> B[B, E]:
        """Order results by these columns ASC or DESC order"""

    @abstractmethod
    def limit(self, limit: int) -> B[B, E]:
        """Limit results"""

    @abstractmethod
    def offset(self, offset: int) -> B[B, E]:
        """Limit offset"""

    @abstractmethod
    def cache(self, name: str = None) -> B[B, E]:
        """Cache results, None seconds uses cache backend default, 0=forever"""

    @abstractmethod
    def sql(self, method: str = 'select') -> str:
        """Get all SQL queries involved in this query builder"""


class DbQueryBuilder(QueryBuilder[B, E], ABC):
    @abstractmethod
    def table(self, table: Union[str, sa.Table]) -> B[B, E]:
        """Add table (select) statement to query"""

    @abstractmethod
    def select(self, *args) -> B[B, E]:
        """Add select (columns) statment to query"""

    @abstractmethod
    def join(self, table: Union[str, sa.Table], left_where: Union[str, sa.Column, BinaryExpression], right_where: Union[str, sa.Column] = None, alias: str = None, method: str = 'join') -> B[B, E]:
        """Add join (default to INNER) statement to query"""

    @abstractmethod
    def outer_join(self, table: Union[str, sa.Table], left_where: Union[str, sa.Column, BinaryExpression], right_where: Union[str, sa.Column] = None, alias: str = None) -> B[B, E]:
        """Add LEFT OUTER join statement to query"""

    @abstractmethod
    def group_by(self, *args) -> B[B, E]:
        """Add group by statement to query"""

    @abstractmethod
    async def find(self, pk_value: Any) -> RowProxy:
        """Execute query by primary key or custom column and return first row found"""

    @abstractmethod
    async def get(self) -> List[RowProxy]:
        """Execute select query and return all rows found"""

    @abstractmethod
    async def delete(self) -> None:
        """Execute delete query"""

class OrmQueryBuilder(QueryBuilder[B, E], ABC):
    @abstractmethod
    def include(self, *args) -> B[B, E]:
        """Include child relation models"""
        pass

    @abstractmethod
    def filter(self, column: Union[str, BinaryExpression, List[Union[Tuple, BinaryExpression]]], operator: str = None, value: Any = None) -> B[B, E]:
        """Filter child relationship by this AND clause"""
        pass

    @abstractmethod
    def or_filter(self, filters: List[Union[Tuple, BinaryExpression]]) -> B[B, E]:
        """Filter child relationship by this OR clause"""
        pass

    @abstractmethod
    def sort(self, column: Union[str, List[str], List[Tuple], Any], order: str = 'ASC') -> B[B, E]:
        """Sort Many relations only"""
        pass

    @abstractmethod
    def key_by(self, field: str) -> B[B, E]:
        """Key results as a Dictionary by this column"""
        pass

    @abstractmethod
    async def find(self, pk_value: Union[int, str] = None, **kwargs) -> Union[E, None]:
        """Execute query by primary key or custom column and return first row found"""
        pass

    @abstractmethod
    async def get(self) -> Union[List[E], Dict[str, E]]:
        """Execute a select query and return all rows found"""
        pass

    @abstractmethod
    async def delete(self) -> None:
        """Execute delete query"""



# type: ignore
from .composer import composer
from .controller import controller
from .event import event
from .job import job
from .model import model
from .provider import provider
from .routes import routes
from .seeder import seeder
from .service import service
from .table import table



from typing import Dict


def deep_merge(override: Dict, default: Dict, *, merge_lists: bool = False) -> Dict:
    """Deep merge a dictionary.
    Values in override will overwrite values in default
    Other non dict values like str, int, list, set will be completely overwritten and not merged.
    This pur function .copy and return a new Dict, it does not transform in place
    """
    override = override.copy()
    default = default.copy()
    def merge(override: Dict, default: Dict) -> Dict:
        for key, value in override.items():
            if isinstance(value, dict):
                # If dict, recurse into values
                node = default.setdefault(key, {})
                merge(value, node)
            elif type(value) == list and merge_lists:
                # If type is list and merge_lists == True, then extend lists
                default[key] = list(set(default[key] + value))  # Sets are unique
            else:
                # If not dict (str, int, array...) override entire value
                # This means arrays will be completely overwritten, not appended
                # Which is accurate for my needs
                default[key] = value
    merge(override, default)
    return default



import os

# If using installed version of asyncclick
#import asyncclick as click
#from asyncclick.termui import _ansi_colors, _ansi_reset_all

# If using local copy of asyncclick
from . import asyncclick as click
from .asyncclick.termui import _ansi_colors, _ansi_reset_all



# MIT License

# Copyright (c) 2016 Roman Tonkonozhko

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# See https://github.com/uvicore/click-help-colors



class HelpColorsException(Exception):
    pass


def _colorize(text, color=None):
    if not color or "NO_COLOR" in os.environ:
        return text
    try:
        return '\033[%dm' % (_ansi_colors[color]) + text + _ansi_reset_all
    except KeyError:
        raise HelpColorsException('Unknown color %r' % color)


def _extend_instance(obj, cls):
    """Apply mixin to a class instance after creation"""
    base_cls = obj.__class__
    base_cls_name = obj.__class__.__name__
    obj.__class__ = type(base_cls_name, (cls, base_cls), {})


class HelpColorsFormatter(click.HelpFormatter):
    def __init__(self, headers_color=None, options_color=None,
                 options_custom_colors=None, *args, **kwargs):
        self.headers_color = headers_color
        self.options_color = options_color
        self.options_custom_colors = options_custom_colors
        super(HelpColorsFormatter, self).__init__(*args, **kwargs)

    def _pick_color(self, option_name):
        opt = option_name.split()[0]
        if (self.options_custom_colors and
                (opt in self.options_custom_colors.keys())):
            return self.options_custom_colors[opt]
        else:
            return self.options_color

    def write_usage(self, prog, args='', prefix='Usage: '):
        colorized_prefix = _colorize(prefix, color=self.headers_color)
        super(HelpColorsFormatter, self).write_usage(prog, args, prefix=colorized_prefix)

    def write_heading(self, heading):
        colorized_heading = _colorize(heading, color=self.headers_color)
        super(HelpColorsFormatter, self).write_heading(colorized_heading)

    def write_dl(self, rows, **kwargs):
        colorized_rows = [(_colorize(row[0], self._pick_color(row[0])), row[1]) for row in rows]
        super(HelpColorsFormatter, self).write_dl(colorized_rows, **kwargs)


class HelpColorsMixin(object):
    def __init__(self, help_headers_color=None, help_options_color=None, help_options_custom_colors=None, *args, **kwargs):
        self.help_headers_color = help_headers_color
        self.help_options_color = help_options_color
        self.help_options_custom_colors = help_options_custom_colors
        super(HelpColorsMixin, self).__init__(*args, **kwargs)

    def get_help(self, ctx):
        formatter = HelpColorsFormatter(
            width=ctx.terminal_width,
            max_width=ctx.max_content_width,
            headers_color=self.help_headers_color,
            options_color=self.help_options_color,
            options_custom_colors=self.help_options_custom_colors)
        self.format_help(ctx, formatter)
        return formatter.getvalue().rstrip('\n')


class HelpColorsGroup(HelpColorsMixin, click.Group):
    def __init__(self, *args, **kwargs):
        super(HelpColorsGroup, self).__init__(*args, **kwargs)

    def command(self, *args, **kwargs):
        kwargs.setdefault('cls', HelpColorsCommand)
        kwargs.setdefault('help_headers_color', self.help_headers_color)
        kwargs.setdefault('help_options_color', self.help_options_color)
        kwargs.setdefault('help_options_custom_colors', self.help_options_custom_colors)
        return super(HelpColorsGroup, self).command(*args, **kwargs)

    def group(self, *args, **kwargs):
        kwargs.setdefault('cls', HelpColorsGroup)
        kwargs.setdefault('help_headers_color', self.help_headers_color)
        kwargs.setdefault('help_options_color', self.help_options_color)
        kwargs.setdefault('help_options_custom_colors', self.help_options_custom_colors)
        return super(HelpColorsGroup, self).group(*args, **kwargs)


class HelpColorsCommand(HelpColorsMixin, click.Command):
    def __init__(self, *args, **kwargs):
        super(HelpColorsCommand, self).__init__(*args, **kwargs)


class HelpColorsMultiCommand(HelpColorsMixin, click.MultiCommand):
    def __init__(self, *args, **kwargs):
        super(HelpColorsMultiCommand, self).__init__(*args, **kwargs)

    def resolve_command(self, ctx, args):
        cmd_name, cmd, args[1:] = super(HelpColorsMultiCommand, self).resolve_command(ctx, args)

        if not isinstance(cmd, HelpColorsMixin):
            if isinstance(cmd, click.Group):
                _extend_instance(cmd, HelpColorsGroup)
            if isinstance(cmd, click.Command):
                _extend_instance(cmd, HelpColorsCommand)

        if not getattr(cmd, 'help_headers_color', None):
            cmd.help_headers_color = self.help_headers_color
        if not getattr(cmd, 'help_options_color', None):
            cmd.help_options_color = self.help_options_color
        if not getattr(cmd, 'help_options_custom_colors', None):
            cmd.help_options_custom_colors = self.help_options_custom_colors

        return cmd_name, cmd, args[1:]



from __future__ import annotations

import operator as operators
from copy import copy
from typing import Any, Dict, Generic, List, Tuple, TypeVar, Union
from uvicore.support.hash import sha1

import sqlalchemy as sa
from sqlalchemy.sql.expression import BinaryExpression
from sqlalchemy.engine.result import Row as RowProxy

import uvicore
from uvicore.database.builder import QueryBuilder, Join
from uvicore.support.dumper import dd, dump
from uvicore.contracts import DbQueryBuilder as BuilderInterface

B = TypeVar("B")  # Builder Type (DbQueryBuilder or OrmQueryBuilder)
E = TypeVar("E")  # Entity Model

@uvicore.service()
class DbQueryBuilder(Generic[B, E], QueryBuilder[B, E], BuilderInterface[B, E]):
    """Database Query Builder"""

    def __init__(self, connection: str):
        self._conn = connection
        super().__init__()

    def table(self, table: Union[str, sa.Table]) -> B[B, E]:
        """Add table (select) statement to query"""
        if type(table) == str:
            self.query.table = uvicore.db.table(table, self._connection())
        else:
            self.query.table = table

        if self.query.table is None:
            raise Exception('Table {} not found.  Are you missing a prefix?  Use uvicore.db.tablename() for proper prefixing.'.format(table))
        return self

    def select(self, *args) -> B[B, E]:
        """Add select (columns) statment to query"""
        for column in args:
            self.query.selects.append(column)
        return self

    def join(self, table: Union[str, sa.Table], left_where: Union[str, sa.Column, BinaryExpression], right_where: Union[str, sa.Column] = None, alias: str = None, method: str = 'join') -> B[B, E]:
        """Add join (default to INNER) statement to query"""
        # Get table and tablename
        conn = self._connection()
        if type(table) == str:
            if '.' in table: conn, table = tuple(table.split('.'))
            table = uvicore.db.table(table, conn)
        tablename = str(table.name)

        # Get left, right and onclause expressions
        left = None
        right = None
        if type(left_where) == BinaryExpression:
            onclause = left_where
        else:
            left = self._column(left_where)
            right = self._column(right_where)
            onclause = left.sacol == right.sacol

        # Set alias to tablename if not defined
        if not alias: alias = tablename

        # Add new Join() expression
        self.query.joins.append(Join(table, tablename, left, right, onclause, alias, method))
        return self

    def outer_join(self, table: Union[str, sa.Table], left_where: Union[str, sa.Column, BinaryExpression], right_where: Union[str, sa.Column] = None, alias: str = None) -> B[B, E]:
        """Add LEFT OUTER join statement to query"""
        self.join(table=table, left_where=left_where, right_where=right_where, method='outerjoin', alias=alias)
        return self

    def group_by(self, *args) -> B[B, E]:
        """Add group by statement to query"""
        for group_by in args:
            self.query.group_by.append(group_by)
        return self

    async def find(self, pk_value: Union[int, str] = None, **kwargs) -> RowProxy:
        """Execute query by primary key or custom column and return first row found"""
        if pk_value:
            # Assume column is PK .find(1234)
            column = self._pk()
            value = pk_value
        elif kwargs:
            # Pass in the column to find .find(email='xyz')
            column = [x for x in kwargs.keys()][0]
            value = [x for x in kwargs.values()][0]

        # Add in where on PK
        self.where(column, value)

        # Get results based on query results
        results = await self.get()

        # Return one record or None
        if results: return results[0]
        return None

    async def get(self) -> List[RowProxy]:
        """Execute select query and return all rows found"""

        # Build select query
        # Use copy() insteadd of .copy() or nested tables tables get deepcopied and SA sees
        # them as multiple tables thworing a 1066, "Not unique table/alias: 'posts'"
        query, saquery = self._build_query('select', copy(self.query))  # do NOT use .copy()

        # Detect caching
        cache = self.query.cache
        if cache:
            prefix = 'uvicore.database/'
            if cache.get('key') is None:
                # No cache name specified, automatically build unique based on queries
                cache['key'] = prefix + query.hash(
                    hash_type='sha1',
                    package='uvicore.database',
                    connection=self._conn,
                )
            else:
                cache['key'] = prefix + cache.get('key')

        # Get cache store from query builder, if None, uses default store
        if cache and await uvicore.cache.store(cache.get('store')).has(cache.get('key')):
            # Cache found, use cached results
            #dump('DB FROM CACHE')
            results = await uvicore.cache.store(cache.get('store')).get(cache.get('key'))
        else:
            # Execute query
            #dump('DB FROM DB')
            results = await uvicore.db.fetchall(saquery, connection=self._connection())

            # Add to cache if desired
            if cache: await uvicore.cache.store(cache.get('store')).put(cache.get('key'), results, seconds=cache.get('seconds'))

        return results

    async def delete(self) -> None:
        """Execute delete query"""

        # Build SQLAlchemy delete query
        query, saquery = self._build_query('delete', copy(self.query))

        # Execute query
        await uvicore.db.execute(saquery, connection=self._connection())

    async def update(self, **kwargs) -> None:
        """Execute update query"""

        # Build SQLAlchemy delete query
        query, saquery = self._build_query('update', copy(self.query))

        # Add in values
        saquery = saquery.values(**kwargs)

        # Execute query
        await uvicore.db.execute(saquery, connection=self._connection())
